/*
**	TerminalPanel.c
**
**	Editing panel for terminal configuration
**
**	Copyright  1990-1995 by Olaf `Olsen' Barthel
**		All Rights Reserved
*/

#include "termGlobal.h"

enum	{	GAD_EMULATIONTYPE=1, GAD_EMULATIONLIBRARY, GAD_COLUMNS, GAD_LINES,
		GAD_KEYMAP, GAD_TEXTFONT, GAD_FONTMODE, GAD_EMULATIONTASK, GAD_IBMFONT,
		GAD_USE, GAD_CANCEL, GAD_DEFAULT, GAD_PAGER, GAD_PAGEGROUP
	};

STATIC LONG	LocalMaxColumns,
		LocalMaxLines;

STATIC STRPTR __saveds
ShowColumn(struct Gadget *SomeGadget,WORD Level)
{
	STATIC UBYTE Buffer[10];

	if(Level == LocalMaxColumns + 1)
		strcpy(Buffer,LocaleString(MSG_TERMINALPANEL_MAXIMUM_TXT));
	else
		SPrintf(Buffer,LocaleBase ? "%9lD" : "%9ld",Level);

	return(Buffer);
}

STATIC STRPTR __saveds
ShowLine(struct Gadget *SomeGadget,WORD Level)
{
	STATIC UBYTE Buffer[10];

	if(Level == LocalMaxLines + 1)
		strcpy(Buffer,LocaleString(MSG_TERMINALPANEL_MAXIMUM_TXT));
	else
		SPrintf(Buffer,LocaleBase ? "%9lD" : "%9ld",Level);

	return(Buffer);
}

	/* SelectKeymap(UBYTE *Name,struct Window *ParentWindow):
	 *
	 *	Select a new keymap file using the asl.library file requester.
	 */

STATIC BYTE __regargs
SelectKeymap(UBYTE *Name,struct Window *ParentWindow)
{
	struct FileRequester	*AslFileRequest;
	struct TagItem		 DimensionTags[5];
	UBYTE			*File,*Temp;
	BYTE			 NewMap = FALSE;

	File = Name;

	strcpy(SharedBuffer,File);

	Temp = PathPart(SharedBuffer);

	Temp[0] = 0;

	if(!SharedBuffer[0])
	{
		APTR OldPtr = ThisProcess -> pr_WindowPtr;
		BPTR FileLock;

		ThisProcess -> pr_WindowPtr = (APTR)-1;

		if(FileLock = Lock("KEYMAPS:",ACCESS_READ))
		{
			strcpy(SharedBuffer,"KEYMAPS:");

			UnLock(FileLock);
		}
		else
			strcpy(SharedBuffer,"Devs:Keymaps");

		ThisProcess -> pr_WindowPtr = OldPtr;
	}

	File = FilePart(File);

	if(AslFileRequest = (struct FileRequester *)AllocAslRequestTags(ASL_FileRequest,
		ASLFR_Window,		ParentWindow,
		ASLFR_InitialFile,	File,
		ASLFR_InitialDrawer,	SharedBuffer,
		ASLFR_TitleText,	LocaleString(MSG_TERMINALPANEL_SELECT_KEYMAP_TXT),
		ASLFR_Flags1,		FILF_NEWIDCMP,
		ASLFR_PrivateIDCMP,	TRUE,
		ASLFR_PositiveText,	LocaleString(MSG_GLOBAL_SELECT_TXT),
		ASLFR_PrivateIDCMP,	TRUE,
	TAG_MORE,GetDimensionTags(ParentWindow,DimensionTags)))
	{
		if(AslRequest(AslFileRequest,NULL))
		{
			STRPTR FileName;

			if(AslFileRequest -> fr_NumArgs > 1 && AslFileRequest -> fr_ArgList)
				FileName = AslFileRequest -> fr_ArgList -> wa_Name;
			else
				FileName = AslFileRequest -> fr_File;

			if(FileName[0])
			{
				PutDimensionTags(ParentWindow,AslFileRequest -> fr_LeftEdge,AslFileRequest -> fr_TopEdge,AslFileRequest -> fr_Width,AslFileRequest -> fr_Height);

				strcpy(SharedBuffer,AslFileRequest -> fr_Drawer);

				if(!AddPart(SharedBuffer,FileName,MAX_FILENAME_LENGTH))
					strcpy(SharedBuffer,FileName);

				if(Stricmp(SharedBuffer,Name))
				{
					strcpy(Name,SharedBuffer);

					NewMap = TRUE;
				}
			}
		}

		FreeAslRequest(AslFileRequest);
	}

	return(NewMap);
}

BYTE __regargs
TerminalPanel(struct Configuration *LocalConfig,BYTE *Default)
{
	STATIC LONG PageTitleID[5] =
	{
		MSG_V36_1020,
		MSG_V36_1021,
		MSG_V36_1024,
		MSG_V36_1025,
		-1
	};

	STATIC BYTE		 InitialPage = 0;

	struct LayoutHandle	*Handle;
	BYTE			 UseIt = FALSE;

	SaveConfig(LocalConfig,PrivateConfig);

	if(Handle = LT_CreateHandleTags(Window -> WScreen,
		LH_LocaleHook,	&LocaleHook,
		LH_ExitFlush,	FALSE,
	TAG_DONE))
	{
		struct Window	*PanelWindow;
		LONG		 LocalColumns,
				 LocalLines;
		UBYTE		 FontName[MAX_FILENAME_LENGTH],
				 IBMName[MAX_FILENAME_LENGTH];

		BuildFontName(FontName,PrivateConfig -> TerminalConfig -> TextFontName,PrivateConfig -> TerminalConfig -> TextFontHeight);

		BuildFontName(IBMName,PrivateConfig -> TerminalConfig -> IBMFontName,PrivateConfig -> TerminalConfig -> IBMFontHeight);

		if(PrivateConfig != Config)
		{
			LocalMaxColumns	= 132;
			LocalMaxLines	= 64;
		}
		else
		{
			LONG	Width,
				Height;

			if(Screen)
			{
				struct Rectangle DisplayClip;

				Width	= PrivateConfig -> ScreenConfig -> DisplayWidth;
				Height	= PrivateConfig -> ScreenConfig -> DisplayHeight;

				QueryOverscan(PrivateConfig -> ScreenConfig -> DisplayMode,&DisplayClip,PrivateConfig -> ScreenConfig -> OverscanType);

				if(!Width)
					Width = DisplayClip . MaxX - DisplayClip . MinX + 1;

				if(!Height)
					Height = DisplayClip . MaxY - DisplayClip . MinY + 1;
			}
			else
			{
				Width	= Window -> WScreen -> Width - Window -> WScreen -> WBorLeft - Window -> WScreen -> WBorRight;
				Height	= Window -> WScreen -> Height - Window -> WScreen -> WBorTop - Window -> WScreen -> WBorBottom - Window -> WScreen -> Font -> ta_YSize - 1;

				if(!StatusWindow)
					Height -= StatusDisplayHeight;

				if(ChatMode)
					Height -= UserFontHeight + 2;
			}

			if(WindowWidth > Width)
				Width = WindowWidth;

			if(WindowHeight > Height)
				Height = WindowHeight;

			LocalMaxColumns	= Width  / TextFontWidth;
			LocalMaxLines	= Height / TextFontHeight;

			if(LocalMaxColumns < 20)
				LocalMaxColumns = 20;

			if(LocalMaxLines < 20)
				LocalMaxLines = 20;
		}

		LT_New(Handle,
			LA_Type,	VERTICAL_KIND,
		TAG_DONE);
		{
			LT_New(Handle,
				LA_Type,VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	TAB_KIND,
					LA_ID,		GAD_PAGER,
					LA_LabelID,	MSG_XFER_PAGE_TXT,
					LACY_LabelTable,PageTitleID,
					LACY_AutoPageID,GAD_PAGEGROUP,
					LACY_TabKey,	TRUE,
					LA_BYTE,	&InitialPage,
					LATB_FullWidth,	TRUE,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
				LA_ID,		GAD_PAGEGROUP,
				LAGR_ActivePage,InitialPage,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	VERTICAL_KIND,
				TAG_DONE);
				{
					LT_New(Handle,
						LA_Type,	CYCLE_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_EMULATION_GAD,
						LA_ID,		GAD_EMULATIONTYPE,
						LA_BYTE,	&PrivateConfig -> TerminalConfig -> EmulationMode,
						LACY_FirstLabel,MSG_TERMINALPANEL_EMULATION_ANSI_VT102_TXT,
						LACY_LastLabel,	MSG_TERMINALPANEL_EMULATION_HEX_TXT,
					TAG_DONE);

					LT_New(Handle,
						LA_Type,	STRING_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_EMULATION_NAME_GAD,
						LA_Chars,	20,
						LA_ID,		GAD_EMULATIONLIBRARY,
						LA_STRPTR,	PrivateConfig -> TerminalConfig -> EmulationFileName,
						GTST_MaxChars,	MAX_FILENAME_LENGTH - 1,
						LAST_Picker,	TRUE,
						GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode != EMULATION_EXTERNAL,
					TAG_DONE);

					LT_New(Handle,
						LA_Type,	CHECKBOX_KIND,
						LA_LabelID,	MSG_EMULATIONPANEL_USE_TERMINAL_TASK_TXT,
						LA_BYTE,	&PrivateConfig -> TerminalConfig -> UseTerminalTask,
						LA_ID,		GAD_EMULATIONTASK,
						GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
					TAG_DONE);

					LT_EndGroup(Handle);
				}

				LT_New(Handle,
					LA_Type,	VERTICAL_KIND,
				TAG_DONE);
				{
					LT_New(Handle,
						LA_Type,	CYCLE_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_BELL_GAD,
						LA_BYTE,	&PrivateConfig -> TerminalConfig -> BellMode,
						LACY_FirstLabel,MSG_TERMINALPANEL_BELL_IGNORE_TXT,
						LACY_LastLabel,	MSG_TERMINALPANEL_BELL_DEFAULT_TXT,
					TAG_DONE);

					if(PrivateConfig -> TerminalConfig -> NumColumns < 20)
						LocalColumns = LocalMaxColumns + 1;
					else
						LocalColumns = PrivateConfig -> TerminalConfig -> NumColumns;

					LT_New(Handle,
						LA_Type,		SliderType,
						LA_LabelID,		MSG_TERMINALPANEL_NUMCOLUMNS_GAD,
						LA_ID,			GAD_COLUMNS,
						LA_LONG,		&LocalColumns,
						GTSL_Min,		20,
						GTSL_Max,		LocalMaxColumns + 1,
						GTSL_DispFunc,		ShowColumn,
						GTSL_LevelFormat,	"%s",
					TAG_DONE);

					if(PrivateConfig -> TerminalConfig -> NumLines < 20)
						LocalLines = LocalMaxLines + 1;
					else
						LocalLines = PrivateConfig -> TerminalConfig -> NumLines;

					LT_New(Handle,
						LA_Type,		SliderType,
						LA_LabelID,		MSG_TERMINALPANEL_NUMLINES_GAD,
						LA_LONG,		&LocalLines,
						LA_ID,			GAD_LINES,
						GTSL_Min,		20,
						GTSL_Max,		LocalMaxLines + 1,
						GTSL_DispFunc,		ShowLine,
						GTSL_LevelFormat,	"%s",
					TAG_DONE);

					LT_New(Handle,
						LA_Type,	STRING_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_KEYMAP_FILE_GAD,
						LA_Chars,	20,
						LA_ID,		GAD_KEYMAP,
						LA_STRPTR,	PrivateConfig -> TerminalConfig -> KeyMapFileName,
						GTST_MaxChars,	MAX_FILENAME_LENGTH - 1,
						LAST_Picker,	TRUE,
					TAG_DONE);

					LT_EndGroup(Handle);
				}

				LT_New(Handle,
					LA_Type,	VERTICAL_KIND,
				TAG_DONE);
				{
					LT_New(Handle,
						LA_Type,	TEXT_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_TEXT_FONT_GAD,
						LA_Chars,	20,
						LA_ID,		GAD_TEXTFONT,
						GTTX_Text,	FontName,
						GTTX_Border,	TRUE,
						LATX_Picker,	TRUE,
						LATX_LockSize,	TRUE,
						GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
					TAG_DONE);

					LT_New(Handle,
						LA_Type,	TEXT_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_IBM_FONT_GAD,
						LA_Chars,	20,
						LA_ID,		GAD_IBMFONT,
						GTTX_Text,	IBMName,
						GTTX_Border,	TRUE,
						LATX_Picker,	TRUE,
						LATX_LockSize,	TRUE,
						GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
					TAG_DONE);

					LT_New(Handle,
						LA_Type,	CYCLE_KIND,
						LA_LabelID,	MSG_TERMINALPANEL_FONT_GAD,
						LA_ID,		GAD_FONTMODE,
						LA_BYTE,	&PrivateConfig -> TerminalConfig -> FontMode,
						LACY_FirstLabel,MSG_TERMINALPANEL_FONT_STANDARD_TXT,
						LACY_LastLabel,	MSG_TERMINALPANEL_FONT_PC_RAW,
						GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
					TAG_DONE);

					LT_EndGroup(Handle);
				}

				LT_New(Handle,
					LA_Type,	HORIZONTAL_KIND,
				TAG_DONE);
				{
					LT_New(Handle,
						LA_Type,	VERTICAL_KIND,
					TAG_DONE);
					{
						LT_New(Handle,
							LA_Type,	CYCLE_KIND,
							LA_LabelID,	MSG_TERMINALPANEL_SEND_CR_GAD,
							LA_BYTE,	&PrivateConfig -> TerminalConfig -> SendCR,
							GTCY_Labels,	EOL_Labels,
						TAG_DONE);

						LT_New(Handle,
							LA_Type,	CYCLE_KIND,
							LA_LabelID,	MSG_TERMINALPANEL_SEND_LF_GAD,
							LA_BYTE,	&PrivateConfig -> TerminalConfig -> SendLF,
							GTCY_Labels,	EOL_Labels,
						TAG_DONE);

						LT_EndGroup(Handle);
					}

					LT_New(Handle,
						LA_Type,	VERTICAL_KIND,
					TAG_DONE);
					{
						LT_New(Handle,
							LA_Type,	YBAR_KIND,
						TAG_DONE);

						LT_EndGroup(Handle);
					}

					LT_New(Handle,
						LA_Type,	VERTICAL_KIND,
					TAG_DONE);
					{
						LT_New(Handle,
							LA_Type,	CYCLE_KIND,
							LA_LabelID,	MSG_TERMINALPANEL_RECEIVE_CR_GAD,
							LA_BYTE,	&PrivateConfig -> TerminalConfig -> ReceiveCR,
							GTCY_Labels,	EOL_Labels,
						TAG_DONE);

						LT_New(Handle,
							LA_Type,	CYCLE_KIND,
							LA_LabelID,	MSG_TERMINALPANEL_RECEIVE_LF_GAD,
							LA_BYTE,	&PrivateConfig -> TerminalConfig -> ReceiveLF,
							GTCY_Labels,	EOL_Labels,
						TAG_DONE);

						LT_EndGroup(Handle);
					}

					LT_EndGroup(Handle);
				}

				LT_EndGroup(Handle);
			}

			LT_New(Handle,
				LA_Type,	VERTICAL_KIND,
			TAG_DONE);
			{
				LT_New(Handle,LA_Type,XBAR_KIND,LAXB_FullSize,TRUE,TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_New(Handle,LA_Type,HORIZONTAL_KIND,
				LAGR_SameSize,	TRUE,
				LAGR_Spread,	TRUE,
			TAG_DONE);
			{
				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_GLOBAL_USE_GAD,
					LA_ID,		GAD_USE,
					LABT_ReturnKey,	TRUE,
					LABT_ExtraFat,	TRUE,
				TAG_DONE);

				if(Default)
				{
					LT_New(Handle,
						LA_Type,	BUTTON_KIND,
						LA_LabelID,	MSG_GLOBAL_DEFAULT_GAD,
						LA_ID,		GAD_DEFAULT,
						LABT_ExtraFat,	TRUE,
					TAG_DONE);
				}

				LT_New(Handle,
					LA_Type,	BUTTON_KIND,
					LA_LabelID,	MSG_GLOBAL_CANCEL_GAD,
					LA_ID,		GAD_CANCEL,
					LABT_EscKey,	TRUE,
					LABT_ExtraFat,	TRUE,
				TAG_DONE);

				LT_EndGroup(Handle);
			}

			LT_EndGroup(Handle);
		}

		if(PanelWindow = LT_Build(Handle,
			LAWN_TitleID,		MSG_TERMINALPANEL_TERMINAL_PREFERENCES_TXT,
			LAWN_IDCMP,		IDCMP_CLOSEWINDOW,
			LAWN_HelpHook,		&GuideHook,
			LAWN_Parent,		Window,
			WA_DepthGadget,		TRUE,
			WA_CloseGadget,		TRUE,
			WA_DragBar,		TRUE,
			WA_RMBTrap,		TRUE,
			WA_Activate,		TRUE,
			WA_SimpleRefresh,	TRUE,
		TAG_DONE))
		{
			struct IntuiMessage	*Message;
			BOOLEAN			 Done = FALSE;
			ULONG			 MsgClass;
			struct Gadget		*MsgGadget;
			UBYTE			 DummyBuffer[MAX_FILENAME_LENGTH];

			GuideContext(CONTEXT_TERMINAL);

			PushWindow(PanelWindow);

			LT_ShowWindow(Handle,TRUE);

			do
			{
				if(Wait(PORTMASK(PanelWindow -> UserPort) | SIG_BREAK) & SIG_BREAK)
					break;

				while(Message = (struct IntuiMessage *)LT_GetIMsg(Handle))
				{
					MsgClass	= Message -> Class;
					MsgGadget	= (struct Gadget *)Message -> IAddress;

					LT_ReplyIMsg(Message);

					if(MsgClass == IDCMP_CLOSEWINDOW)
						Done = TRUE;

					if(MsgClass == IDCMP_GADGETUP)
					{
						switch(MsgGadget -> GadgetID)
						{
							case GAD_USE:

								LT_UpdateStrings(Handle);

								PrivateConfig -> TerminalConfig -> NumColumns	= LocalColumns % (LocalMaxColumns + 1);
								PrivateConfig -> TerminalConfig -> NumLines	= LocalLines % (LocalMaxLines + 1);

								if(!PrivateConfig -> TerminalConfig -> EmulationFileName[0] && PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL)
									PrivateConfig -> TerminalConfig -> EmulationMode = EMULATION_ANSIVT100;

								UseIt = Done = TRUE;
								break;

							case GAD_EMULATIONTYPE:

								LT_SetAttributes(Handle,GAD_TEXTFONT,
									GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
								TAG_DONE);

								LT_SetAttributes(Handle,GAD_IBMFONT,
									GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
								TAG_DONE);

								LT_SetAttributes(Handle,GAD_FONTMODE,
									GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
								TAG_DONE);

								LT_SetAttributes(Handle,GAD_EMULATIONLIBRARY,
									GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode != EMULATION_EXTERNAL,
								TAG_DONE);

								LT_SetAttributes(Handle,GAD_EMULATIONTASK,
									GA_Disabled,	PrivateConfig -> TerminalConfig -> EmulationMode == EMULATION_EXTERNAL,
								TAG_DONE);

								break;

							case GAD_CANCEL:

								Done = TRUE;
								break;

							case GAD_DEFAULT:

								UseIt = *Default = TRUE;

								Done = TRUE;
								break;
						}
					}

					if(MsgClass == IDCMP_IDCMPUPDATE)
					{
						switch(MsgGadget -> GadgetID)
						{
							case GAD_EMULATIONLIBRARY:

								LT_LockWindow(PanelWindow);

								strcpy(DummyBuffer,(STRPTR)LT_GetAttributes(Handle,GAD_EMULATIONLIBRARY,TAG_DONE));
								strcpy(SharedBuffer,DummyBuffer);

								if(PickFile(PanelWindow,"Libs:","xem#?.library",LocaleString(MSG_TERMINALPANEL_SELECT_EMULATOR_TXT),SharedBuffer,NT_LIBRARY))
								{
									if(Stricmp(SharedBuffer,DummyBuffer))
										LT_SetAttributes(Handle,GAD_EMULATIONLIBRARY,GTST_String,SharedBuffer,TAG_DONE);
								}

								LT_ShowWindow(Handle,TRUE);

								LT_UnlockWindow(PanelWindow);
								break;

							case GAD_KEYMAP:

								LT_LockWindow(PanelWindow);

								strcpy(DummyBuffer,(STRPTR)LT_GetAttributes(Handle,GAD_KEYMAP,TAG_DONE));

								if(SelectKeymap(DummyBuffer,PanelWindow))
									LT_SetAttributes(Handle,GAD_KEYMAP,GTST_String,SharedBuffer,TAG_DONE);

								LT_UnlockWindow(PanelWindow);
								break;

							case GAD_TEXTFONT:

								if(PickFont(PanelWindow,PrivateConfig -> TerminalConfig -> TextFontName,&PrivateConfig -> TerminalConfig -> TextFontHeight,TRUE))
								{
									BuildFontName(FontName,PrivateConfig -> TerminalConfig -> TextFontName,PrivateConfig -> TerminalConfig -> TextFontHeight);

									LT_SetAttributes(Handle,GAD_TEXTFONT,
										GTTX_Text,	FontName,
									TAG_DONE);
								}

								break;

							case GAD_IBMFONT:

								if(PickFont(PanelWindow,PrivateConfig -> TerminalConfig -> IBMFontName,&PrivateConfig -> TerminalConfig -> IBMFontHeight,TRUE))
								{
									BuildFontName(IBMName,PrivateConfig -> TerminalConfig -> IBMFontName,PrivateConfig -> TerminalConfig -> IBMFontHeight);

									LT_SetAttributes(Handle,GAD_IBMFONT,
										GTTX_Text,	IBMName,
									TAG_DONE);
								}

								break;
						}
					}
				}
			}
			while(!Done);

			PopWindow();
		}

		LT_DeleteHandle(Handle);
	}

	if(UseIt)
		SwapConfig(LocalConfig,PrivateConfig);
	else
		SaveConfig(LocalConfig,PrivateConfig);

	return(UseIt);
}
